# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import random
import weakref
import concurrent.futures
import numbers
import functools
import collections.abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Mapping, MutableMapping)
import hashlib
import weakref
import dataclasses
import pathlib
import tempfile

import more_itertools
import numpy as np

import marley.utils
from marley import gamey
from marley.jamswank.swanking import SimpleField, SavvyField, JamId, SwankDatabase
from .base import Observation, Action, Story
from .policing import Policy, QPolicy
from . import utils

DEFAULT_MAX_BATCH_SIZE = 128
DEFAULT_MAX_PAST_MEMORY_SIZE = 1_000

class MustDefineCustomModel(NotImplementedError):
    pass


@dataclasses.dataclass(frozen=True)
class ModelSpec:
    create_model: Callable
    observation_neural_dtype: np.dtype
    action_n_neurons: int

    def __call__(self) -> 'keras.Model':
        return self.create_model(self.observation_neural_dtype, self.action_n_neurons)


def _serialize_model(model: 'keras.Model') -> bytes:
    with marley.utils.create_temp_folder() as temp_folder:
        path = pathlib.Path(temp_folder) / 'model.h5'
        model.save_weights(path, save_format='h5')
        return path.read_bytes()

def _deserialize_to_model(model: 'keras.Model',
                          serialized_model: collections.abc.ByteString) -> None:
    with marley.utils.create_temp_folder() as temp_folder:
        path = pathlib.Path(temp_folder) / 'model.h5'
        path.write_bytes(serialized_model)
        model.load_weights(path)




class ModelJockey(collections.abc.Mapping):
    def __init__(self, model_spec: ModelSpec, max_size: int = 5) -> None:
        self.model_spec = model_spec
        self.weak_model_tracker = weakref.WeakValueDictionary()
        self.model_to_serialized_model = utils.WeakKeyIdentityDict()
        self.serialized_model_to_model = marley.utils.LruDict(max_size)

    @property
    def max_size(self) -> int:
        return self.serialized_model_to_model.max_size

    def __getitem__(self, serialized_model: bytes) -> 'keras.Model':
        try:
            return self.serialized_model_to_model[serialized_model]
        except KeyError:
            model = self.model_spec()
            _deserialize_to_model(model, serialized_model)

            self.weak_model_tracker[id(model)] = model
            self.model_to_serialized_model[model] = serialized_model
            self.serialized_model_to_model[serialized_model] = model
            return model


    def get_random_model(self) -> 'keras.Model':
        model = self.model_spec()
        serialized_model = _serialize_model(model)

        self.weak_model_tracker[id(model)] = model
        self.model_to_serialized_model[model] = serialized_model
        self.serialized_model_to_model[serialized_model] = model
        return model


    def get_random_serialized_model(self) -> 'keras.Model':
        return self.model_to_serialized_model[self.get_random_model()]

    def __iter__(self) -> Iterator[bytes]:
        return iter(self.serialized_model_to_model)

    def __len__(self) -> int:
        return len(self.serialized_model_to_model)


    def serialize_model(self, model: 'keras.Model') -> bytes:
        try:
            return self.model_to_serialized_model[model]
        except AttributeError:
            self.model_to_serialized_model[model] = serialized_model = _serialize_model(model)
            return serialized_model


    def clone_model_and_train(self, serialized_model: bytes,
                              train: Callable[['keras.Model'], None]) -> bytes:
        if len(self) < self.max_size:
            model = self.model_spec()
        else:
            # Todo: Two race condition: Peek and get, and someone using this model elsewhere
            oldest_serialized_model, model = self.serialized_model_to_model.peek_last_item()
            del self.serialized_model_to_model[oldest_serialized_model]
            del self.model_to_serialized_model[model]

        _deserialize_to_model(model, serialized_model)
        train(model)
        new_serialized_model = _serialize_model(model)

        self.weak_model_tracker[id(model)] = model
        self.model_to_serialized_model[model] = new_serialized_model
        self.serialized_model_to_model[new_serialized_model] = model

        return new_serialized_model



class ModelMegaJockey(collections.abc.Mapping):

    def __init__(self):
        self._dict = {}

    def __len__(self) -> int:
        return len(self._dict)

    def __iter__(self) -> Iterator[ModelSpec]:
        return iter(self._dict)

    def __getitem__(self, model_spec: ModelSpec) -> ModelJockey:
        try:
            return self._dict[model_spec]
        except KeyError:
            self._dict[model_spec] = model_jockey = ModelJockey(model_spec)
            return model_jockey


model_mega_jockey = ModelMegaJockey()

class ModelManager(collections.abc.Sequence):
    def __init__(self, model_free_learning_policy: ModelFreeLearningPolicy) -> None:
        self.model_free_learning_policy = model_free_learning_policy

    def __len__(self):
        return len(self.model_free_learning_policy.serialized_models)

    def __getitem__(self, i: Union[int, slice]) -> 'keras.Model':
        if isinstance(i, slice):
            raise NotImplementedError
        assert isinstance(i, int)
        return self.model_free_learning_policy.model_jockey[
            self.model_free_learning_policy.serialized_models[i]
        ]




class ModelFreeLearningPolicy(QPolicy):

    action_type: Type[Action] = SavvyField()
    observation_neural_dtype: np.dtype = SavvyField()
    epsilon: numbers.Real = SimpleField()
    discount: numbers.Real = SimpleField()
    serialized_models: Sequence[bytes] = SavvyField()

    @classmethod
    def create(cls, *, action_type: Optional[Type[Action]] = None,
               observation: Optional[Union[Observation, Type[Observation]]] = None,
               observation_neural_dtype: Optional[np.dtype] = None,
               serialized_models: Optional[Sequence[bytes]] = None, default_n_models: int = 2,
               epsilon: numbers.Real = 0.1, discount: numbers.Real = 0.9) -> None:


        ### Figuring out `action_type`: ############################################################
        #                                                                                          #
        if action_type is None:
            default_action_type = cls.action_type.default_factory()
            assert issubclass(default_action_type, Action)
            action_type = default_action_type
        #                                                                                          #
        ### Finished figuring out `action_dtype`. ##################################################


        ### Figuring out `observation_neural_dtype`: ###############################################
        #                                                                                          #
        if observation_neural_dtype is not None:
            assert observation is None
        elif observation is not None:
            observation_neural_dtype = observation.neural_dtype
        elif hasattr(cls, 'observation_type'):
            observation_neural_dtype = cls.observation_type.neural_dtype
        else:
            assert isinstance(cls.observation_neural_dtype, np.dtype) is not None
            # Defined as class attribute
        #                                                                                          #
        ### Finished figuring out `observation_neural_dtype`. ######################################

        ### Figuring out `serialized_models`: ######################################################
        #                                                                                          #
        if serialized_models is None:
            model_jockey = model_mega_jockey[
                ModelSpec(
                    cls.create_model, observation_neural_dtype=observation_neural_dtype,
                    action_n_neurons=action_type.n_neurons
                )
            ]
            serialized_models = tuple(model_jockey.get_random_serialized_model()
                                      for _ in range(default_n_models))
        else:
            serialized_models = tuple(serialized_models)
        #                                                                                          #
        ### Finished figuring out `serialized_models`: #############################################

        return cls(
            action_type=action_type, observation_neural_dtype=observation_neural_dtype,
            epsilon=epsilon, discount=discount, serialized_models=serialized_models
        )



    def __init__(self, *, action_type: Type[Action], observation_neural_dtype: np.dtype,
                 epsilon: numbers.Real, discount: numbers.Real,
                 serialized_models: Sequence[bytes], jam_id: Optional[JamId] = None,
                 jam_index: Optional[int] = None,
                 swank_database: Optional[SwankDatabase] = None) -> None:
        QPolicy.__init__(
            self, action_type=action_type, observation_neural_dtype=observation_neural_dtype,
            epsilon=epsilon, discount=discount, serialized_models=serialized_models, jam_id=jam_id,
            jam_index=jam_index, swank_database=swank_database

        )
        self.fingerprint = hashlib.sha512(b''.join(self.serialized_models)).hexdigest()[:6]
        self.models = ModelManager(self)


    @property
    def _model_kwargs(self):
        return dict(observation_neural_dtype=self.observation_neural_dtype,
                    action_n_neurons=self.action_type.n_neurons)

    @property
    def _model_spec(self):
        return ModelSpec(self.create_model, **self._model_kwargs)

    def _get_qs_for_observations_uncached(self, observations: Sequence[Observation] = None) -> \
                                                               Tuple[Mapping[Action, numbers.Real]]:
        if not observations:
            return ()
        observation_neurals = [observation.neural for observation in observations]
        if observation_neurals:
            assert utils.is_structured_array(observation_neurals[0])
        input_array = np.concatenate(observation_neurals)
        model = random.choice(self.models)
        prediction_output = self.predict(model, input_array)
        actions = tuple(self.action_type)
        return tuple(
            {action: q for action, q in dict(zip(actions, output_row)).items()
             if (action in observation.legal_actions)}
            for observation, output_row in zip(observations, prediction_output)
        )

    def train(self, narratives: Sequence[gamey.Narrative], *, n_epochs: int = 10,
              max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
              max_past_memory_size: int = DEFAULT_MAX_PAST_MEMORY_SIZE) -> ModelFreeLearningPolicy:

        clone_kwargs = self._Swank__field_values

        serialized_models = self.serialized_models
        train_model = lambda model: self._train_model(
            model, narratives, n_epochs=n_epochs,
            other_model=self.model_jockey[serialized_models[-1]],
            max_batch_size=max_batch_size, max_past_memory_size=max_past_memory_size
        )
        serialized_trained_model = self.model_jockey.clone_model_and_train(
            serialized_models[0], train_model
        )
        clone_kwargs['serialized_models'] = serialized_models[1:] + (serialized_trained_model,)
        return type(self)(**clone_kwargs)



    def get_next_action(self, observation: Observation) -> Action:
        epsilon = self.epsilon
        if (epsilon > 0) and (epsilon == 1 or epsilon > random.random()):
            # The verbose condition above is an optimized version of `if epsilon > random.random():`
            return random.choice(observation.legal_actions)
        else:
            q_map = self.get_qs_for_observation(observation) # This is cached.
            return max(q_map, key=q_map.__getitem__)




    def get_observation_v(self, observation: Observation,
                           epsilon: Optional[numbers.Real] = None) -> numbers.Real:
        if epsilon is None:
            epsilon = self.epsilon
        q_map = self.get_qs_for_observation(observation)
        return np.average(
            (
                max(q_map.values()),
                np.average(tuple(q_map.values()))
            ),
            weights=(1 - epsilon, epsilon)
        )

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.fingerprint}>'

    @property
    def model_jockey(self) -> ModelJockey:
        return model_mega_jockey[self._model_spec]


    @staticmethod
    def create_model(observation_neural_dtype: np.dtype, action_n_neurons: int) -> 'keras.Model':
        from tensorflow import keras
        if tuple(observation_neural_dtype.fields) != ('sequential_input',):
            raise MustDefineCustomModel
        model = keras.models.Sequential(
            layers=(
                keras.layers.Input(
                    shape=observation_neural_dtype['sequential_input'].shape,
                    name='sequential_input'
                ),
                keras.layers.Dense(
                    256, activation='relu',
                ),
                keras.layers.Dropout(rate=0.1),
                keras.layers.Dense(
                    256, activation='relu',
                ),
                keras.layers.Dropout(rate=0.1),
                keras.layers.Dense(
                    256, activation='relu',
                ),
                keras.layers.Dropout(rate=0.1),
                keras.layers.Dense(
                    action_n_neurons # activation='relu'
                ),
            ),
        )
        model.compile(optimizer='rmsprop', loss='mse', metrics=['accuracy'])

        return model


    def predict(self, model: 'keras.Model', input_array: np.ndarray) -> np.ndarray:
        return model.predict({name: input_array[name] for name in input_array.dtype.names})


    def _train_model(self, model: 'keras.Model', narratives: Sequence[gamey.Narrative], *,
                     other_model: 'keras.Model', n_epochs: int,
                     max_batch_size: int = DEFAULT_MAX_BATCH_SIZE,
                     max_past_memory_size: int = DEFAULT_MAX_PAST_MEMORY_SIZE) -> None:

        ### Getting a random selection of stories to train on: #####################################
        #                                                                                          #
        past_memory = utils.ChainSpace(map(reversed, reversed(narratives)))
        foo = min(max_past_memory_size, len(past_memory))
        batch_size = min(max_batch_size, foo)
        indices = tuple(sorted(random.sample(range(foo), batch_size)))
        stories = tuple(past_memory[index] for index in indices)
        #                                                                                          #
        ### Finished getting a random selection of stories to train on. ############################

        ### Initializing arrays: ###################################################################
        #                                                                                          #
        old_observation_neural_array = np.zeros(
            (batch_size,), dtype=self.observation_neural_dtype
        )
        action_neural_array = np.zeros(
            (batch_size, self.action_type.n_neurons), dtype=bool
        )
        reward_array = np.zeros(batch_size)
        new_observation_neural_array = np.zeros(
            (batch_size,), dtype=self.observation_neural_dtype
        )
        are_not_end_array = np.zeros(batch_size, dtype=bool)

        for i, story in enumerate(stories):
            story: Story
            old_observation_neural_array[i] = story.old_observation.neural
            action_neural_array[i] = story.action.to_neural()
            reward_array[i] = story.reward
            new_observation_neural_array[i] = story.new_observation.neural
            are_not_end_array[i] = not story.new_observation.state.is_end

        #                                                                                          #
        ### Finished initializing arrays. ##########################################################

        prediction = self.predict(
            model,
            np.concatenate((old_observation_neural_array, new_observation_neural_array))
        )
        wip_q_values, new_q_values = np.split(prediction, 2)

        if other_model is model:
            new_other_q_values = new_q_values
        else:
            new_other_q_values = self.predict(
                other_model,
                new_observation_neural_array
            )


        action_indices = np.dot(action_neural_array,
                                range(self.action_type.n_neurons)).astype(np.int32)
        batch_index = np.arange(batch_size, dtype=np.int32)
        wip_q_values[batch_index, action_indices] = (
            reward_array + self.discount * are_not_end_array *
            new_other_q_values[np.arange(new_q_values.shape[0]),
                               np.argmax(new_q_values, axis=1)]
        )


        fit_arguments = {
            'x': {name: old_observation_neural_array[name] for name
                  in old_observation_neural_array.dtype.names},
            'y': wip_q_values,
            'verbose': 0,
            'epochs': n_epochs,
        }

        model.fit(**fit_arguments)



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
import numbers
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict)


from . import int_crowding
from .int_crowding import IntCrowd, Interval
from .gaining import Gain, ThinGain, FatGain

def clamp_non_negative(number: numbers.Number) -> numbers.Number:
    if number >= 0:
        return number
    else:
        return 0



class Weight(abc.ABC):
    _number: int

    def __init__(self, number: int = 0) -> None:
        self._number = number

    def _reduce(self) -> tuple:
        return (type(self), self._number)

    def __eq__(self, other: Any) -> bool:
        return (type(other) == type(self)) and (self._reduce() == other._reduce())

    def __hash__(self) -> int:
        return hash(self._reduce())

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._number})'


    @abc.abstractmethod
    def get_unblocked_child_gain(self, desired_child_gain: Gain,
                                 finished_parent_gain: Gain) -> Gain:
        raise NotImplementedError

    @abc.abstractmethod
    def get_desired_parent_gain(self, desired_child_gain: Gain, parent_job: Job) -> Gain:
        raise NotImplementedError


class CalfWeight(Weight):
    def __init__(self, offset: int = 0) -> None:
        assert isinstance(offset, int)
        self.offset = offset
        Weight.__init__(self, offset)

    def get_unblocked_child_gain(self, desired_child_gain: Gain,
                                 finished_parent_gain: Gain) -> Gain:
        from .jobbing import ThinJob, FatJob
        is_thin_gain = isinstance(desired_child_gain, ThinGain)
        assert is_thin_gain == isinstance(finished_parent_gain, ThinGain)
        if is_thin_gain:
            return ThinGain(
                desired_child_gain.job,
                (desired_child_gain and finished_parent_gain)
            )
        else:
            assert isinstance(desired_child_gain, FatGain)
            assert isinstance(finished_parent_gain, FatGain)
            main_intervals = tuple(
                Interval(interval.sliceful_tuple[:-1] +
                        (slice(clamp_non_negative(interval.sliceful_tuple[-1].start - self.offset),
                               clamp_non_negative(interval.sliceful_tuple[-1].stop - self.offset)),))
                for interval in finished_parent_gain.int_crowd.intervals
            )
            if self.offset < 0:
                trimmed_int_crowd = desired_child_gain.int_crowd.trim_inclusive()
                bonus_intervals = tuple(
                    Interval(interval.as_tuple() + ((0, -self.offset),))
                    for interval in trimmed_int_crowd.intervals
                )
            else:
                bonus_intervals = ()
            intervals = main_intervals + bonus_intervals
            return desired_child_gain.job.create_gain(desired_child_gain.int_crowd & intervals)

    def get_desired_parent_gain(self, desired_child_gain: Gain, parent_job: Job) -> Gain:
        from .jobbing import ThinJob
        is_thin_gain = isinstance(desired_child_gain, ThinGain)
        assert is_thin_gain == isinstance(parent_job, ThinJob)
        if is_thin_gain:
            return ThinGain(parent_job, desired_child_gain.is_finished)
        else:
            assert isinstance(desired_child_gain, FatGain)
            assert isinstance(parent_job, FatJob)
            return parent_job.create_gain(
                IntCrowd(
                    Interval((
                        interval.sliceful_tuple[:-1] +
                         (slice(clamp_non_negative(interval.sliceful_tuple[-1].start + self.offset),
                               clamp_non_negative(interval.sliceful_tuple[-1].stop + self.offset)),)
                    ))
                     for interval in desired_child_gain.int_crowd.intervals
                )
            )


class TadpoleWeight(Weight):
    def __init__(self, amount: int = 0) -> None:
        assert isinstance(amount, int)
        self.amount = amount
        Weight.__init__(self, amount)

    def get_unblocked_child_gain(self, desired_child_gain: Gain,
                                 finished_parent_gain: FatGain) -> Gain:
        assert isinstance(finished_parent_gain, FatGain)
        if isinstance(desired_child_gain, ThinGain):
            assert finished_parent_gain.job.dimensions == 1
            return ThinGain(desired_child_gain.job,
                            desired_child_gain and (Interval[0 : self.amount] in
                                                    finished_parent_gain.int_crowd))
        else:
            return desired_child_gain.job.create_gain(
                IntCrowd(
                    point for point in desired_child_gain.int_crowd if
                    Interval(point.as_tuple() + ((0, self.amount),))
                    in finished_parent_gain.int_crowd
                )
            )


    def get_desired_parent_gain(self, desired_child_gain: Gain, parent_job: FatJob) -> Gain:
        assert isinstance(parent_job, FatJob)
        if isinstance(desired_child_gain, ThinGain):
            assert parent_job.dimensions == 1
            return parent_job.create_gain(
                IntCrowd(Interval[0 : self.amount]) if desired_child_gain else None
            )
        else:
            return parent_job.create_gain(
                IntCrowd(
                    Interval(interval.as_tuple() + ((0, self.amount),)) for interval in
                    desired_child_gain.int_crowd.intervals
                )
            )

from .jobbing import Job, ThinJob, FatJob



# Copyright 2020 Ram Rachum and collaborators.
# This program is distributed under the MIT license.

from __future__ import annotations

import abc
from typing import (Iterable, Union, Optional, Tuple, Any, Iterator, Type, FrozenSet,
                    Sequence, Callable, Hashable, Mapping, TypeVar, Dict, Set)
import itertools
import collections
import dataclasses

import networkx as nx

from .jobbing import Job
from .weighting import Weight, CalfWeight, TadpoleWeight

@dataclasses.dataclass(order=True, frozen=True)
class Relationship:
    parent_job: Job
    child_job: Job
    weight: Weight



class Wedge:
    def __init__(self) -> None:
        self.graph = nx.DiGraph()

    def _add_edge(self, parent_job: Job, child_job: Job,
                 weight: Union[CalfWeight, TadpoleWeight]) -> None:
        from .jobbing import ThinJob, FatJob
        assert isinstance(parent_job, Job)
        assert isinstance(child_job, Job)
        assert isinstance(weight, (CalfWeight, TadpoleWeight))
        assert child_job.get_parent_job_to_weight()[parent_job] == weight
        if isinstance(weight, CalfWeight):
            assert isinstance(parent_job, ThinJob) == isinstance(child_job, ThinJob)
            assert parent_job.dimensions == child_job.dimensions
        else:
            assert isinstance(weight, TadpoleWeight)
            assert isinstance(parent_job, FatJob)
            assert parent_job.dimensions == child_job.dimensions + 1
        try:
            existing_edge_info = self.graph.edges[(parent_job, child_job)]
        except KeyError:
            self.graph.add_edge(parent_job, child_job, weight=weight)
        else:
            existing_weight = existing_edge_info['weight']
            assert weight == existing_weight


    def add_jobs(self, jobs: Union[Job, Iterable[Job]]) -> None:
        jobs: set = {jobs} if isinstance(jobs, Job) else set(jobs)
        jobs_to_process = set(jobs)
        jobs_processed = set()
        while jobs_to_process:
            job: Job = jobs_to_process.pop()
            self.graph.add_node(job)
            for parent_job, weight in job.get_parent_job_to_weight().items():
                self._add_edge(parent_job, job, weight)
                if parent_job not in jobs_processed:
                    jobs_to_process.add(parent_job)
            jobs_processed.add(job)

    def get_child_jobs(self, jobs: Union[Job, Iterable[Job]], *,
                       include_self: bool = False) -> Set[Job]:
        jobs: set = {jobs} if isinstance(jobs, Job) else set(jobs)
        child_jobs = set(itertools.chain.from_iterable(self.graph[job] for job in jobs))
        return (child_jobs | jobs) if include_self else child_jobs

    # def get_child_job_to_weight(self, job: Job) -> Dict[Job, Weight]:
        # child_jobs = {}
        # for child, d in self.graph.reverse()[job].items():
            # child_jobs[child] = d['weight']
        # return child_jobs


    def iterate_predecessor_jobs(self, jobs: Union[Job, Iterable[Job]], *,
                                 include_self: bool = False) -> Iterator[Job]:
        return self._iterate(jobs,
                             lambda job: tuple(job.get_parent_job_to_weight()),
                             include_self=include_self)


    def iterate_successor_jobs(self, jobs: Union[Job, Iterable[Job]], *,
                               include_self: bool = False) -> Iterator[Job]:
        return self._iterate(jobs,
                             lambda job: tuple(self.graph[job]),
                             include_self=include_self)


    def _iterate(self, jobs: Union[Job, Iterable[Job]], step: Callable[[Job], Iterable[Job]], *,
                 include_self: bool = False):
        jobs: set = {jobs} if isinstance(jobs, Job) else set(jobs)
        if include_self:
            yield from jobs
        all_jobs = set(jobs)
        while jobs:
            jobs = set(itertools.chain.from_iterable(map(step, jobs))) - all_jobs
            all_jobs |= jobs
            yield from jobs






